# Copyright 2021, Guillermo Adrián Molina
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import deque
from itertools import islice

from zcm import zcm_config

# TODO: use from prettytable import PrettyTable ?

def print_table(table, header=True, truncate=True, separation=2, identation=0, page_size=25):
    if page_size <= 0:
        page_size = None
    i = iter(table)
    ask_for_more = False
    while True:
        page = tuple(islice(i, 0, page_size))
        if len(page):
            if ask_for_more:
                answer = input('Do you want to see more? (Y/n) ')
                if answer and answer.upper()[0] == 'N':
                    return
            print_table_page(page, header, truncate, separation, identation)
            ask_for_more = True
        else:
            return


def print_table_page(page, header=True, truncate=True, separation=2, identation=0):
    MAX_COLUMN_LENGTH = zcm_config['max_column_length']
    if len(page) == 0:
        return

    columns = []
    # initialize columns from first row's keys
    for key in page[0]:
        columns.append({
            'key': key,
            'tittle': key.upper(),
            'length': len(key)
        })

    # adjust columns lenghts to max record sizes
    for column in columns:
        for row in page:
            value = str(row[column['key']]).replace('\t', ' ')
            row[column['key']] = value
            column['length'] = max(column['length'], len(value))

    if truncate:
        for column in columns:
            column['length'] = min(column['length'], MAX_COLUMN_LENGTH)

    separation_string = ' ' * separation

    # print headers
    if header:
        strings = [''] * identation if identation > 0 else []
        for column in columns:
            str_format = '{:%s}' % str(column['length'])
            strings.append(str_format.format(column['tittle']))
        print(separation_string.join(strings))

    for row in page:
        strings = [''] * identation if identation > 0 else []
        for column in columns:
            value = row[column['key']]
            if truncate and len(value) > MAX_COLUMN_LENGTH:
                value = value[:MAX_COLUMN_LENGTH-3] + '...'
            str_format = '{:%s}' % str(column['length'])
            strings.append(str_format.format(value))
        print(separation_string.join(strings))

def format_bytes(size):
    # 2**10 = 1024
    power = 2**10
    n = 0
    power_labels = {0: 'B', 1: 'KB', 2: 'MB', 3: 'GB', 4: 'TB', 5: 'PB'}
    while size > power:
        size /= power
        n += 1
    return '{:.2f} {:s}'.format(size, power_labels[n])


def print_info(data):
    for key, value in data.items():
        print('%s: %s' % (key, value))



# Copyright 2021, Guillermo Adrián Molina
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import argparse

from zcm.api.manager import Manager
from zcm.lib.helpers import check_one_or_more, check_positive


class Activate:
    name = 'activate'
    aliases = []

    @staticmethod
    def init_parser(parent_subparsers):
        parent_parser = argparse.ArgumentParser(add_help=False)
        parser = parent_subparsers.add_parser(Activate.name,
                                              parents=[parent_parser],
                                              aliases=Activate.aliases,
                                              formatter_class=argparse.ArgumentDefaultsHelpFormatter,
                                              description='activate an clone',
                                              help='activate an clone')
        parser.add_argument('-m', '--max-newer',
                            type=check_positive,
                            help='Do not activate if there will be more than <max-newer> newer clones')
        parser.add_argument('-M', '--max-older',
                            type=check_positive,
                            help='Do not activate if there will be more than <max-older> older clones')
        parser.add_argument('-t', '--max-total',
                            type=check_one_or_more,
                            help='Do not clone if there are <max-total> clones')
        parser.add_argument('-a', '--auto-remove',
                            action='store_true',
                            help='Remove clones if maximum limits excedeed')
        parser.add_argument('path',
                            metavar='filesystem|path',
                            help='zfs filesystem or path of ZCM')
        parser.add_argument('id',
                            help='clone id to activate')

    def __init__(self, options):
        manager = Manager(options.path)
        manager.activate(options.id, options.max_newer,
                                 options.max_older, options.max_total, options.auto_remove)
        if not options.quiet:
            print('Activated clone ' + options.id)



# Copyright 2021, Guillermo Adrián Molina
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import argparse
import logging

from zcm import __version__
from zcm.cli.activate import Activate
from zcm.cli.clone import Clone
from zcm.cli.destroy import Destroy
from zcm.cli.difference import Difference
from zcm.cli.information import Information
from zcm.cli.initialize import Initialize
from zcm.cli.list import List
from zcm.cli.remove import Remove
from zcm.exceptions import ZCMException

log = logging.getLogger(__name__)


def set_log_level(log_level):
    if log_level == 'none':
        logging.disable(logging.CRITICAL)
    else:
        levels = {
            'debug': logging.DEBUG,
            'info': logging.INFO,
            'warn': logging.WARNING,
            'error': logging.ERROR,
            'critical': logging.CRITICAL
        }
        logging.basicConfig(level=levels[log_level])


class CustomFormatter(argparse.ArgumentDefaultsHelpFormatter,
                      argparse.RawDescriptionHelpFormatter):
    pass


class CLI:
    commands = [Initialize, Information, List, Clone, Activate, Difference, Remove, Destroy]

    def __init__(self):
        parser = argparse.ArgumentParser(
            formatter_class=CustomFormatter,
            description='Tool to manage ZFS clones with history metadata')
        parser.add_argument('-V', '--version',
                            help='Print version information and quit',
                            action='version',
                            version='%(prog)s version ' + __version__)
        parser.add_argument('-l', '--log-level',
                            help='Set the logging level ("debug"|"info"|"warn"|"error"|"fatal")',
                            choices=[
                                'debug',
                                'info',
                                'warn',
                                'error',
                                'critical',
                                'none'
                            ],
                            metavar='LOG_LEVEL',
                            default='none')
        parser.add_argument('-D', '--debug',
                            help='Enable debug mode',
                            action='store_true')
        parser.add_argument('-q', '--quiet',
                            help='Enable quiet mode',
                            action='store_true')

        subparsers = parser.add_subparsers(
            dest='command',
            metavar='COMMAND',
            required=True)

        for command in CLI.commands:
            command.init_parser(subparsers)

        options = parser.parse_args()

        set_log_level(options.log_level)

        if options.debug:
            import debugpy
            debugpy.listen(('0.0.0.0', 5678))
            log.info("Waiting for IDE to attach...")
            debugpy.wait_for_client()

        try:
            for command in self.commands:
                if options.command == command.name or options.command in command.aliases:
                    command(options)
                    break
        except ZCMException as e:
            log.error(e.message)
            print(e.message)
            exit(-1)


def main():
    CLI()


if __name__ == '__main__':
    main()



# Copyright 2020, Guillermo Adrián Molina
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import io
import logging
import pathlib
import subprocess
from datetime import datetime

log = logging.getLogger(__name__)


class ZFSError(Exception):
    def __init__(self, message="ZFS Error"):
        super().__init__()
        self.message = message


def get_cmd(command,  arguments, options):
    cmd = ['/usr/sbin/zfs', command]
    if options is not None:
        for option in options:
            cmd += ['-o', option]
    if arguments is not None:
        cmd += arguments
    log.debug('Running command: "' + ' '.join(cmd) + '"')
    return cmd


def _zfs(command,  arguments=None, options=None, stdout=None):
    cmd = get_cmd(command, arguments, options)
    return subprocess.Popen(cmd, stdout=stdout)


def zfs(command,  arguments=None, options=None):
    cmd = get_cmd(command, arguments, options)
    process = subprocess.run(cmd, capture_output=True, text=True)
    if process.stdout:
        for line in iter(process.stdout.splitlines()):
            log.info(line)
    if process.stderr:
        for line in iter(process.stderr.splitlines()):
            if process.returncode == 0:
                log.warning(line)
            else:
                log.error(line)
    if process.returncode != 0:
        raise ZFSError(process.stderr)
    return process.stdout


def zfs_create(zfs_name, parent=None, mountpoint=None, compression=None, recursive=False, zcm_path=None):
    filesystem = zfs_name
    if parent is None:
        last_index = zfs_name.rfind('/')
        parent = filesystem[:last_index]
        zfs_name = filesystem[last_index+1:]
    else:
        filesystem = '%s/%s' % (parent, zfs_name)

    if recursive:
        zfs_path = ''
        is_zpool = True
        for zfs_fs in parent.split('/'):
            zfs_path += '/' if zfs_path else ''
            zfs_path += zfs_fs
            if not zfs_is_filesystem(zfs_path):
                if is_zpool:
                    return None
                zfs('create', [zfs_path])

            is_zpool = False

    options = []
    if compression is not None:
        options.append('compression=' + compression)
    if mountpoint is not None:
        options.append('mountpoint=' + str(mountpoint))
    if zcm_path is not None:
        options.append('zfs_clone_manager:path=' + str(zcm_path))
    if len(options) == 0:
        options = None
    zfs('create', [filesystem], options)
    return filesystem


def zfs_clone(zfs_name, snapshot, parent=None, mountpoint=None):
    if not isinstance(zfs_name, str):
        log.error('The ZFS clone name must be provided')
    if not isinstance(snapshot, str):
        log.error('The ZFS snapshot must be provided')
    filesystem = zfs_name
    if parent is not None:
        filesystem = parent + '/' + zfs_name
    options = []
    if mountpoint is not None:
        options.append('mountpoint=' + str(mountpoint))
    zfs('clone', [snapshot, filesystem], options)
    return filesystem


def zfs_set(zfs_name, readonly=None, mountpoint=None, zcm_path=None):
    result = []
    if readonly is not None:
        option = 'readonly=' + ('on' if readonly else 'off')
        result.append(zfs('set', [option, zfs_name]))
    if mountpoint is not None:
        result.append(zfs('set', ['mountpoint=' + str(mountpoint), zfs_name]))
    if zcm_path is not None:
        result.append(zfs(
            'set', ['zfs_clone_manager:path=' + str(zcm_path), zfs_name]))
    return '\n'.join(result)


def zfs_inherit(zfs_name, property_name):
    return zfs('inherit', [property_name, zfs_name])


def value_convert(property_name, value):
    if value == 'on':
        return True
    if value == 'off':
        return False
    if value == '-':
        return None
    if property_name in ['mountpoint', 'zfs_clone_manager:path']:
        return pathlib.Path(value)
    if property_name in ['creation', 'st_ctim', 'mtime', 'atime', 'crtime']:
        try:
            return datetime.fromtimestamp(float(value))
        except ValueError:
            pass
    try:
        return int(value)
    except ValueError:
        return value


def zfs_get(zfs_name, property_name):
    if property_name == 'all':
        raise NotImplementedError()
    cmd = ['/usr/sbin/zfs', 'get', '-Hp', property_name, zfs_name]
    with open('/dev/null', 'w') as dev_null:
        log.debug('Running command: "' + ' '.join(cmd) + '"')
        output = subprocess.check_output(cmd, stderr=dev_null)
        value = output.decode('utf-8').split('\t')[2]
        return value_convert(property_name, value)
    return None


def zfs_snapshot(zfs_name, filesystem, recursive=False):
    arguments = []
    if recursive:
        arguments.append('-r')
    snapshot = filesystem + '@' + zfs_name
    arguments.append(snapshot)
    zfs('snapshot', arguments)
    return snapshot


def zfs_destroy(zfs_name, recursive=False, synchronous=True):
    arguments = []
    if recursive:
        arguments.append('-r')
    if synchronous:
        arguments.append('-s')
    arguments.append(zfs_name)
    return zfs('destroy', arguments)


def zfs_promote(zfs_name):
    return zfs('promote', [zfs_name])


def zfs_send(last_snapshot, target_file_path, first_snapshot=None, recursive=False):
    arguments = []
    if recursive:
        arguments.append('-R')
    if first_snapshot is not None:
        arguments += ['-I', first_snapshot]
    arguments.append(last_snapshot)
    with open(target_file_path, 'wb') as target_file:
        return _zfs('send', arguments, stdout=target_file)


def zfs_list(zfs_name=None, zfs_type=None, recursive=False,
             properties=['name', 'used', 'avail', 'refer', 'mountpoint']):
    cmd = ['/usr/sbin/zfs', 'list', '-Hp']
    if recursive:
        cmd.append('-r')
    if zfs_type is not None and zfs_type in ['all', 'filesystem',
                                             'snapshot', 'volume']:
        cmd += ['-t', zfs_type]
    if properties is not None:
        cmd += ['-o', ','.join(properties)]
    if zfs_name is not None:
        cmd.append(zfs_name)
    try:
        with open('/dev/null', 'w') as dev_null:
            filesystems = []
            log.debug('Running command: "' + ' '.join(cmd) + '"')
            output = subprocess.check_output(cmd, stderr=dev_null)
            for line in output.decode('utf-8').strip().split('\n'):
                values = line.split('\t')
                filesystem = {}
                for property_name, value in zip(properties, values):
                    filesystem[property_name] = value_convert(
                        property_name, value)
                filesystems.append(filesystem)
            return filesystems
    except:
        return []


def zfs_exists(zfs_name):
    filesystems = zfs_list(zfs_name, zfs_type='all', properties=['name'])
    return len(filesystems) == 1


def zfs_is_filesystem(zfs_name):
    try:
        return zfs_get(zfs_name, 'type') == 'filesystem'
    except:
        return False


def zfs_is_snapshot(zfs_name):
    try:
        return zfs_get(zfs_name, 'type') == 'snapshot'
    except:
        return False


def zfs_diff(zfs_name, origin_snapshot=None, include_file_types=False, recursive=False):
    # Implemented as generator, in case it is too big
    file_types = {
        'F': 'file',
        '/': 'directory',
        'B': 'device',
        '>': 'door',
        '|': 'fifo',
        '@': 'link',
        'P': 'portal',
        '=': 'socket'
    }
    change_types = {
        '+': 'Added',
        '-': 'Removed',
        'M': 'Modified',
        'R': 'Renamed'
    }

    arguments = ['-H', '-t']
    if include_file_types:
        arguments.append('-F')
    if recursive:
        arguments.append('-r')
    if origin_snapshot is None:
        arguments.append('-E')
    else:
        arguments.append(origin_snapshot)
    arguments.append(zfs_name)
    mountpoint = None
    if not '@' in zfs_name:
        mountpoint = zfs_get(zfs_name, 'mountpoint')
    process = _zfs('diff', arguments, stdout=subprocess.PIPE)
    for line in io.TextIOWrapper(process.stdout, encoding="utf-8"):
        records = line.strip().split('\t')
        data = {}
        file = pathlib.Path(records[-1])
        if mountpoint is not None:
            data['mountpoint'] = mountpoint
            file = file.relative_to(mountpoint)
        data['date'] = value_convert('st_ctim', records[0])
        data['change'] = change_types[records[1]]
        data['file'] = file
        if include_file_types:
            data['file_type'] = file_types[records[2]]
        yield data


def zfs_rename(original_zfs_name, new_zfs_name):
    arguments = [original_zfs_name, new_zfs_name]
    return zfs('rename', arguments)


def zfs_mount(zfs_name):
    return zfs('mount', [zfs_name])


def zfs_unmount(zfs_name):
    return zfs('unmount', [zfs_name])



# Copyright 2021, Guillermo Adrián Molina
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import json

from zcm.api.manager import Manager
from zcm.lib.print import format_bytes, print_table


class List:
    name = 'list'
    aliases = ['ls']

    @staticmethod
    def init_parser(parent_subparsers):
        parent_parser = argparse.ArgumentParser(add_help=False)
        parser = parent_subparsers.add_parser(List.name,
                                              parents=[parent_parser],
                                              aliases=List.aliases,
                                              formatter_class=argparse.ArgumentDefaultsHelpFormatter,
                                              description='List clones',
                                              help='List clones')
        parser.add_argument('-j', '--json',
                            action='store_true',
                            help='Output a JSON object')
        parser.add_argument('-T', '--no-trunc',
                            help='Don\'t truncate output',
                            action='store_true')
        parser.add_argument('-H', '--no-header',
                            help='Don\'t show header line(s)',
                            action='store_true')
        parser.add_argument('-P', '--page-size',
                            help='If list is longer than <page-size>, ask for more to continue in <page-size> intervals.\
                                Enter 0 to avoid pagination',
                            type=int,
                            default=25)
        parser.add_argument('path',
                            nargs='*',
                            metavar='filesystem|path',
                            help='zfs filesystem or path to show')

    def __init__(self, options):
        table = []
        managers = []
        if options.path:
            managers = [Manager(path) for path in options.path]
        else:
            managers = Manager.get_managers()
        if options.json:
            clones = []
            for manager in managers:
                for clone in manager.clones:
                    clones.append(clone.to_dictionary())
            print(json.dumps(clones, indent=4))
        else:
            for manager in managers:
                for clone in manager.clones:
                    table.append({
                        'manager': manager.zfs,
                        'a': '*' if manager.active_clone == clone else ' ',
                        'id': clone.id,
                        'clone': clone.zfs,
                        'mountpoint': str(clone.mountpoint),
                        'origin': clone.origin_id if clone.origin_id else '',
                        'date': clone.creation,
                        'size': format_bytes(clone.size)
                    })
            print_table(table, header=(not options.no_header), truncate=(
                not options.no_trunc), page_size=options.page_size)



# Copyright 2021, Guillermo Adrián Molina
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import logging
import random
import string
import subprocess

log = logging.getLogger(__name__)


def check_positive(value, min_value=0):
    ivalue = int(value)
    if ivalue < min_value:
        raise argparse.ArgumentTypeError(
            "%s is an invalid positive int value" % value)
    return ivalue


def check_one_or_more(value):
    return check_positive(value, 1)


# https://stackoverflow.com/questions/2257441/random-string-generation-with-upper-case-letters-and-digits
def id_generator(size=10, chars=string.ascii_uppercase + string.digits):
    return ''.join(random.choice(chars) for _ in range(size))


# shutil.copytree does not copy owner
def copy_directory(source, target):
    if not source.is_dir() or not target.is_dir():
        return -1
    cmd1 = ['tar', 'cf', '-', '.']
    cmd2 = ['tar', 'xf', '-']
    log.debug('Running command: "cd %s; %s | (cd %s; %s)"' % (str(source), ' '.join(cmd1), str(target), ' '.join(cmd2)))
    process1 = subprocess.Popen(cmd1, stdout=subprocess.PIPE, cwd=source)
    process2 = subprocess.Popen(cmd2, stdin=process1.stdout, cwd=target)
    process1.stdout.close()
    stdout, stderr = process2.communicate()
    if stdout:
        for line in iter(stdout.splitlines()):
            log.info(line)
    if stderr:
        for line in iter(stderr.splitlines()):
            if process2.returncode == 0:
                log.warning(line)
            else:
                log.error(line)
    return process2.returncode



# Copyright 2021, Guillermo Adrián Molina
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import argparse

from zcm.api.manager import Manager


def are_you_sure(force, manager):
    if force:
        return True
    print('WARNING!!!!!!!!')
    print('All the filesystems, clones, snapshots and directories associated with %s will be permanently deleted.' % manager.zfs)
    print('This operation is not reversible.')
    answer = input('Do you want to proceed? (yes/NO) ')
    return answer == 'yes'


class Destroy:
    name = 'destroy'
    aliases = []

    @staticmethod
    def init_parser(parent_subparsers):
        parent_parser = argparse.ArgumentParser(add_help=False)
        parser = parent_subparsers.add_parser(Destroy.name,
                                              parents=[parent_parser],
                                              aliases=Destroy.aliases,
                                              formatter_class=argparse.ArgumentDefaultsHelpFormatter,
                                              description='Destroy ZCM on path',
                                              help='Remove all ZCM metadata (filesystems, clones, snapshots and directories) associated with path')
        parser.add_argument('-F', '--force',
                            help='Force destroy without confirmation',
                            action='store_true')
        parser.add_argument('path',
                            metavar='filesystem|path',
                            nargs='+',
                            help='zfs filesystem or path of ZCM')

    def __init__(self, options):
        managers = [ Manager(path) for path in options.path ]
        for manager in managers:
            if are_you_sure(options.force, manager):
                manager.destroy()
                if not options.quiet:
                    print('Destroyed ZCM %s' % manager.zfs)



# Copyright 2021, Guillermo Adrián Molina
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
from datetime import datetime
from pathlib import Path

from zcm.api.manager import Manager
from zcm.exceptions import ZCMError, ZCMException
from zcm.lib.zfs import zfs_exists, zfs_get, zfs_is_filesystem

zfs = 'rpool/my/cool/zfs/directory'
directory = '/my_cool_zfs_directory'


class TestAPI(unittest.TestCase):
    def setUp(self):
        try:
            Manager.initialize_manager(zfs, directory)
        except ZCMError:
            pass
        return super().setUp()

    def tearDown(self):
        try:
            Manager(directory).destroy()
        except ZCMError:
            pass
        return super().tearDown()

    def test_initialize(self):
        with self.assertRaises(ZCMError):
            Manager.initialize_manager(zfs, directory)
        clone = None
        try:
            manager = Manager(directory)
        except ZCMError as e:
            self.fail('Instantiation should not raise exceptions')

        self.assertEqual(manager.path, Path(directory))
        self.assertEqual(manager.zfs, zfs)
        self.assertEqual(manager.active_clone, manager.clones[0])
        self.assertEqual(manager.next_id, '00000001')
        self.assertEqual(len(manager.older_clones), 0)
        self.assertEqual(len(manager.newer_clones), 0)

        filesystem = zfs
        path = Path(directory, '.clones')
        self.assertTrue(zfs_is_filesystem(filesystem))
        self.assertEqual(zfs_get(filesystem, 'mountpoint'), path)
        self.assertTrue(zfs_get(filesystem, 'mounted'))
        self.assertTrue(path.is_dir())

        id = '00000000'
        filesystem = '%s/%s' % (zfs, id)
        path = Path(directory)
        self.assertEqual(manager.clones[0].id, id)
        self.assertEqual(manager.clones[0].zfs, filesystem)
        self.assertIsNone(manager.clones[0].origin)
        self.assertIsNone(manager.clones[0].origin_id)
        self.assertEqual(manager.clones[0].mountpoint, path)
        self.assertIsInstance(manager.clones[0].creation, datetime)
        self.assertTrue(zfs_is_filesystem(filesystem))
        self.assertEqual(zfs_get(filesystem, 'mountpoint'), path)
        self.assertTrue(zfs_get(filesystem, 'mounted'))
        self.assertTrue(path.is_dir())

    def test_create_1(self):
        clone = None
        try:
            manager = Manager(directory)
        except ZCMError as e:
            self.fail('Instantiation should not raise exceptions')
        try:
            manager.clone()
        except ZCMError as e:
            self.fail('Creation should not raise exceptions')

        self.assertEqual(manager.path, Path(directory))
        self.assertEqual(manager.zfs, zfs)
        self.assertEqual(manager.active_clone, manager.clones[0])
        self.assertEqual(manager.next_id, '00000002')
        self.assertEqual(len(manager.older_clones), 0)
        self.assertEqual(len(manager.newer_clones), 1)

        filesystem = zfs
        path = Path(directory, '.clones')
        self.assertTrue(zfs_is_filesystem(filesystem))
        self.assertEqual(zfs_get(filesystem, 'mountpoint'), path)
        self.assertTrue(zfs_get(filesystem, 'mounted'))
        self.assertTrue(path.is_dir())

        id = '00000000'
        filesystem = '%s/%s' % (zfs, id)
        path = Path(directory)
        self.assertEqual(manager.clones[0].id, id)
        self.assertEqual(manager.clones[0].zfs, filesystem)
        self.assertIsNone(manager.clones[0].origin)
        self.assertIsNone(manager.clones[0].origin_id)
        self.assertEqual(manager.clones[0].mountpoint, path)
        self.assertIsInstance(manager.clones[0].creation, datetime)
        self.assertTrue(zfs_is_filesystem(filesystem))
        self.assertEqual(zfs_get(filesystem, 'mountpoint'), path)
        self.assertTrue(zfs_get(filesystem, 'mounted'))
        self.assertTrue(path.is_dir())

        id = '00000001'
        filesystem = '%s/%s' % (zfs, id)
        path = Path(directory, '.clones', id)
        self.assertEqual(manager.clones[1].id, id)
        self.assertEqual(manager.clones[1].zfs, filesystem)
        self.assertEqual(manager.clones[1].origin, '%s/%s@%s' % (zfs, '00000000', id))
        self.assertEqual(manager.clones[1].origin_id, '00000000')
        self.assertEqual(manager.clones[1].mountpoint, path)
        self.assertIsInstance(manager.clones[1].creation, datetime)
        self.assertTrue(zfs_is_filesystem(filesystem))
        self.assertEqual(zfs_get(filesystem, 'mountpoint'), path)
        self.assertTrue(zfs_get(filesystem, 'mounted'))
        self.assertTrue(path.is_dir())

    def test_activate_1(self):
        clone = None
        try:
            manager = Manager(directory)
        except ZCMError as e:
            self.fail('Instantiation should not raise exceptions')
        try:
            manager.clone()
        except ZCMError as e:
            self.fail('Creation should not raise exceptions')
        try:
            manager.activate('00000001')
        except ZCMError as e:
            self.fail('Creation should not raise exceptions')
        with self.assertRaises(ZCMException):
            manager.activate('00000001')
        with self.assertRaises(ZCMError):
            manager.remove('00000001')

        self.assertEqual(manager.path, Path(directory))
        self.assertEqual(manager.zfs, zfs)
        self.assertEqual(manager.active_clone, manager.clones[1])
        self.assertEqual(manager.next_id, '00000002')
        self.assertEqual(len(manager.older_clones), 1)
        self.assertEqual(len(manager.newer_clones), 0)

        filesystem = zfs
        path = Path(directory, '.clones')
        self.assertTrue(zfs_is_filesystem(filesystem))
        self.assertEqual(zfs_get(filesystem, 'mountpoint'), path)
        self.assertTrue(zfs_get(filesystem, 'mounted'))
        self.assertTrue(path.is_dir())

        id = '00000000'
        filesystem = '%s/%s' % (zfs, id)
        path = Path(directory, '.clones', id)
        self.assertEqual(manager.clones[0].id, id)
        self.assertEqual(manager.clones[0].zfs, filesystem)
        self.assertIsNone(manager.clones[0].origin)
        self.assertIsNone(manager.clones[0].origin_id)
        self.assertEqual(manager.clones[0].mountpoint, path)
        self.assertIsInstance(manager.clones[0].creation, datetime)
        self.assertTrue(zfs_is_filesystem(filesystem))
        self.assertEqual(zfs_get(filesystem, 'mountpoint'), path)
        self.assertTrue(zfs_get(filesystem, 'mounted'))
        self.assertTrue(path.is_dir())

        id = '00000001'
        filesystem = '%s/%s' % (zfs, id)
        path = Path(directory)
        self.assertEqual(manager.clones[1].id, id)
        self.assertEqual(manager.clones[1].zfs, filesystem)
        self.assertEqual(manager.clones[1].origin, '%s/%s@%s' % (zfs, '00000000', id))
        self.assertEqual(manager.clones[1].origin_id, '00000000')
        self.assertEqual(manager.clones[1].mountpoint, path)
        self.assertIsInstance(manager.clones[1].creation, datetime)
        self.assertTrue(zfs_is_filesystem(filesystem))
        self.assertEqual(zfs_get(filesystem, 'mountpoint'), path)
        self.assertTrue(zfs_get(filesystem, 'mounted'))
        self.assertTrue(path.is_dir())

    def test_remove_newer_1(self):
        clone = None
        try:
            manager = Manager(directory)
        except ZCMError as e:
            self.fail('Instantiation should not raise exceptions')
        try:
            manager.clone()
        except ZCMError as e:
            self.fail('Creation should not raise exceptions')
        try:
            manager.remove('00000001')
        except ZCMError as e:
            self.fail('Creation should not raise exceptions')

        self.assertEqual(manager.path, Path(directory))
        self.assertEqual(manager.zfs, zfs)
        self.assertEqual(manager.active_clone, manager.clones[0])
        self.assertEqual(manager.next_id, '00000001')
        self.assertEqual(len(manager.older_clones), 0)
        self.assertEqual(len(manager.newer_clones), 0)

        filesystem = zfs
        path = Path(directory, '.clones')
        self.assertTrue(zfs_is_filesystem(filesystem))
        self.assertEqual(zfs_get(filesystem, 'mountpoint'), path)
        self.assertTrue(zfs_get(filesystem, 'mounted'))
        self.assertTrue(path.is_dir())

        id = '00000000'
        filesystem = '%s/%s' % (zfs, id)
        path = Path(directory)
        self.assertEqual(manager.clones[0].id, id)
        self.assertEqual(manager.clones[0].zfs, filesystem)
        self.assertIsNone(manager.clones[0].origin)
        self.assertIsNone(manager.clones[0].origin_id)
        self.assertEqual(manager.clones[0].mountpoint, path)
        self.assertIsInstance(manager.clones[0].creation, datetime)
        self.assertTrue(zfs_is_filesystem(filesystem))
        self.assertEqual(zfs_get(filesystem, 'mountpoint'), path)
        self.assertTrue(zfs_get(filesystem, 'mounted'))
        self.assertTrue(path.is_dir())

        id = '00000001'
        filesystem = '%s/%s' % (zfs, id)
        path = Path(directory, '.clones', id)
        self.assertFalse(zfs_exists(filesystem))
        self.assertFalse(path.exists())

    def test_remove_older_1(self):
        clone = None
        try:
            manager = Manager(directory)
        except ZCMError as e:
            self.fail('Instantiation should not raise exceptions')
        try:
            manager.clone()
        except ZCMError as e:
            self.fail('Creation should not raise exceptions')
        try:
            manager.activate('00000001')
        except ZCMError as e:
            self.fail('Creation should not raise exceptions')
        try:
            manager.remove('00000000')
        except ZCMError as e:
            self.fail('Creation should not raise exceptions')

        self.assertEqual(manager.path, Path(directory))
        self.assertEqual(manager.zfs, zfs)
        self.assertEqual(manager.active_clone, manager.clones[0])
        self.assertEqual(manager.next_id, '00000002')
        self.assertEqual(len(manager.older_clones), 0)
        self.assertEqual(len(manager.newer_clones), 0)

        filesystem = zfs
        path = Path(directory, '.clones')
        self.assertTrue(zfs_is_filesystem(filesystem))
        self.assertEqual(zfs_get(filesystem, 'mountpoint'), path)
        self.assertTrue(zfs_get(filesystem, 'mounted'))
        self.assertTrue(path.is_dir())

        id = '00000000'
        filesystem = '%s/%s' % (zfs, id)
        path = Path(directory, '.clones', id)
        self.assertFalse(zfs_exists(filesystem))
        self.assertFalse(path.exists())

        id = '00000001'
        filesystem = '%s/%s' % (zfs, id)
        path = Path(directory)
        self.assertEqual(manager.clones[0].id, id)
        self.assertEqual(manager.clones[0].zfs, filesystem)
        self.assertIsNone(manager.clones[0].origin)
        self.assertIsNone(manager.clones[0].origin_id)
        self.assertEqual(manager.clones[0].mountpoint, path)
        self.assertIsInstance(manager.clones[0].creation, datetime)
        self.assertTrue(zfs_is_filesystem(filesystem))
        self.assertEqual(zfs_get(filesystem, 'mountpoint'), path)
        self.assertTrue(zfs_get(filesystem, 'mounted'))
        self.assertTrue(path.is_dir())

    def test_create_2(self):
        clone = None
        try:
            manager = Manager(directory)
        except ZCMError as e:
            self.fail('Instantiation should not raise exceptions')
        try:
            manager.clone()
        except ZCMError as e:
            self.fail('Creation should not raise exceptions')
        try:
            manager.clone()
        except ZCMError as e:
            self.fail('Creation should not raise exceptions')

        self.assertEqual(manager.path, Path(directory))
        self.assertEqual(manager.zfs, zfs)
        self.assertEqual(manager.active_clone, manager.clones[0])
        self.assertEqual(manager.next_id, '00000003')
        self.assertEqual(len(manager.older_clones), 0)
        self.assertEqual(len(manager.newer_clones), 2)

        filesystem = zfs
        path = Path(directory, '.clones')
        self.assertTrue(zfs_is_filesystem(filesystem))
        self.assertEqual(zfs_get(filesystem, 'mountpoint'), path)
        self.assertTrue(zfs_get(filesystem, 'mounted'))
        self.assertTrue(path.is_dir())

        id = '00000000'
        filesystem = '%s/%s' % (zfs, id)
        path = Path(directory)
        self.assertEqual(manager.clones[0].id, id)
        self.assertEqual(manager.clones[0].zfs, filesystem)
        self.assertIsNone(manager.clones[0].origin)
        self.assertIsNone(manager.clones[0].origin_id)
        self.assertEqual(manager.clones[0].mountpoint, path)
        self.assertIsInstance(manager.clones[0].creation, datetime)
        self.assertTrue(zfs_is_filesystem(filesystem))
        self.assertEqual(zfs_get(filesystem, 'mountpoint'), path)
        self.assertTrue(zfs_get(filesystem, 'mounted'))
        self.assertTrue(path.is_dir())

        id = '00000001'
        filesystem = '%s/%s' % (zfs, id)
        path = Path(directory, '.clones', id)
        self.assertEqual(manager.clones[1].id, id)
        self.assertEqual(manager.clones[1].zfs, filesystem)
        self.assertEqual(manager.clones[1].origin, '%s/%s@%s' % (zfs, '00000000', id))
        self.assertEqual(manager.clones[1].origin_id, '00000000')
        self.assertEqual(manager.clones[1].mountpoint, path)
        self.assertIsInstance(manager.clones[1].creation, datetime)
        self.assertTrue(zfs_is_filesystem(filesystem))
        self.assertEqual(zfs_get(filesystem, 'mountpoint'), path)
        self.assertTrue(zfs_get(filesystem, 'mounted'))
        self.assertTrue(path.is_dir())

        id = '00000002'
        filesystem = '%s/%s' % (zfs, id)
        path = Path(directory, '.clones', id)
        self.assertEqual(manager.clones[2].id, id)
        self.assertEqual(manager.clones[2].zfs, filesystem)
        self.assertEqual(manager.clones[2].origin, '%s/%s@%s' % (zfs, '00000000', id))
        self.assertEqual(manager.clones[2].origin_id, '00000000')
        self.assertEqual(manager.clones[2].mountpoint, path)
        self.assertIsInstance(manager.clones[1].creation, datetime)
        self.assertTrue(zfs_is_filesystem(filesystem))
        self.assertEqual(zfs_get(filesystem, 'mountpoint'), path)
        self.assertTrue(zfs_get(filesystem, 'mounted'))
        self.assertTrue(path.is_dir())

    def test_activate_2(self):
        clone = None
        try:
            manager = Manager(directory)
        except ZCMError as e:
            self.fail('Instantiation should not raise exceptions')
        try:
            manager.clone()
        except ZCMError as e:
            self.fail('Creation should not raise exceptions')
        try:
            manager.clone()
        except ZCMError as e:
            self.fail('Creation should not raise exceptions')
        try:
            manager.activate('00000001')
        except ZCMError as e:
            self.fail('Creation should not raise exceptions')
        with self.assertRaises(ZCMException):
            manager.activate('00000001')
        with self.assertRaises(ZCMError):
            manager.remove('00000001')

        self.assertEqual(manager.path, Path(directory))
        self.assertEqual(manager.zfs, zfs)
        self.assertEqual(manager.active_clone, manager.clones[1])
        self.assertEqual(manager.next_id, '00000003')
        self.assertEqual(len(manager.older_clones), 1)
        self.assertEqual(len(manager.newer_clones), 1)

        filesystem = zfs
        path = Path(directory, '.clones')
        self.assertTrue(zfs_is_filesystem(filesystem))
        self.assertEqual(zfs_get(filesystem, 'mountpoint'), path)
        self.assertTrue(zfs_get(filesystem, 'mounted'))
        self.assertTrue(path.is_dir())

        id = '00000000'
        filesystem = '%s/%s' % (zfs, id)
        path = Path(directory, '.clones', id)
        self.assertEqual(manager.clones[0].id, id)
        self.assertEqual(manager.clones[0].zfs, filesystem)
        self.assertIsNone(manager.clones[0].origin)
        self.assertIsNone(manager.clones[0].origin_id)
        self.assertEqual(manager.clones[0].mountpoint, path)
        self.assertIsInstance(manager.clones[0].creation, datetime)
        self.assertTrue(zfs_is_filesystem(filesystem))
        self.assertEqual(zfs_get(filesystem, 'mountpoint'), path)
        self.assertTrue(zfs_get(filesystem, 'mounted'))
        self.assertTrue(path.is_dir())

        id = '00000001'
        filesystem = '%s/%s' % (zfs, id)
        path = Path(directory)
        self.assertEqual(manager.clones[1].id, id)
        self.assertEqual(manager.clones[1].zfs, filesystem)
        self.assertEqual(manager.clones[1].origin, '%s/%s@%s' % (zfs, '00000000', id))
        self.assertEqual(manager.clones[1].origin_id, '00000000')
        self.assertEqual(manager.clones[1].mountpoint, path)
        self.assertIsInstance(manager.clones[1].creation, datetime)
        self.assertTrue(zfs_is_filesystem(filesystem))
        self.assertEqual(zfs_get(filesystem, 'mountpoint'), path)
        self.assertTrue(zfs_get(filesystem, 'mounted'))
        self.assertTrue(path.is_dir())

        id = '00000002'
        filesystem = '%s/%s' % (zfs, id)
        path = Path(directory, '.clones', id)
        self.assertEqual(manager.clones[2].id, id)
        self.assertEqual(manager.clones[2].zfs, filesystem)
        self.assertEqual(manager.clones[2].origin, '%s/%s@%s' % (zfs, '00000000', id))
        self.assertEqual(manager.clones[2].origin_id, '00000000')
        self.assertEqual(manager.clones[2].mountpoint, path)
        self.assertIsInstance(manager.clones[2].creation, datetime)
        self.assertTrue(zfs_is_filesystem(filesystem))
        self.assertEqual(zfs_get(filesystem, 'mountpoint'), path)
        self.assertTrue(zfs_get(filesystem, 'mounted'))
        self.assertTrue(path.is_dir())

    def test_remove_newer_2(self):
        clone = None
        try:
            manager = Manager(directory)
        except ZCMError as e:
            self.fail('Instantiation should not raise exceptions')
        try:
            manager.clone()
        except ZCMError as e:
            self.fail('Creation should not raise exceptions')
        try:
            manager.clone()
        except ZCMError as e:
            self.fail('Creation should not raise exceptions')
        try:
            manager.activate('00000001')
        except ZCMError as e:
            self.fail('Creation should not raise exceptions')
        try:
            manager.remove('00000002')
        except ZCMError as e:
            self.fail('Creation should not raise exceptions')

        self.assertEqual(manager.path, Path(directory))
        self.assertEqual(manager.zfs, zfs)
        self.assertEqual(manager.active_clone, manager.clones[1])
        self.assertEqual(manager.next_id, '00000002')
        self.assertEqual(len(manager.older_clones), 1)
        self.assertEqual(len(manager.newer_clones), 0)

        filesystem = zfs
        path = Path(directory, '.clones')
        self.assertTrue(zfs_is_filesystem(filesystem))
        self.assertEqual(zfs_get(filesystem, 'mountpoint'), path)
        self.assertTrue(zfs_get(filesystem, 'mounted'))
        self.assertTrue(path.is_dir())

        id = '00000000'
        filesystem = '%s/%s' % (zfs, id)
        path = Path(directory, '.clones', id)
        self.assertEqual(manager.clones[0].id, id)
        self.assertEqual(manager.clones[0].zfs, filesystem)
        self.assertIsNone(manager.clones[0].origin)
        self.assertIsNone(manager.clones[0].origin_id)
        self.assertEqual(manager.clones[0].mountpoint, path)
        self.assertIsInstance(manager.clones[0].creation, datetime)
        self.assertTrue(zfs_is_filesystem(filesystem))
        self.assertEqual(zfs_get(filesystem, 'mountpoint'), path)
        self.assertTrue(zfs_get(filesystem, 'mounted'))
        self.assertTrue(path.is_dir())

        id = '00000001'
        filesystem = '%s/%s' % (zfs, id)
        path = Path(directory)
        self.assertEqual(manager.clones[1].id, id)
        self.assertEqual(manager.clones[1].zfs, filesystem)
        self.assertEqual(manager.clones[1].origin, '%s/%s@%s' % (zfs, '00000000', id))
        self.assertEqual(manager.clones[1].origin_id, '00000000')
        self.assertEqual(manager.clones[1].mountpoint, path)
        self.assertIsInstance(manager.clones[1].creation, datetime)
        self.assertTrue(zfs_is_filesystem(filesystem))
        self.assertEqual(zfs_get(filesystem, 'mountpoint'), path)
        self.assertTrue(zfs_get(filesystem, 'mounted'))
        self.assertTrue(path.is_dir())

        id = '00000002'
        filesystem = '%s/%s' % (zfs, id)
        path = Path(directory, '.clones', id)
        self.assertFalse(zfs_exists(filesystem))
        self.assertFalse(path.exists())

    def test_remove_older_2(self):
        clone = None
        try:
            manager = Manager(directory)
        except ZCMError as e:
            self.fail('Instantiation should not raise exceptions')
        try:
            manager.clone()
        except ZCMError as e:
            self.fail('Creation should not raise exceptions')
        try:
            manager.clone()
        except ZCMError as e:
            self.fail('Creation should not raise exceptions')
        try:
            manager.activate('00000001')
        except ZCMError as e:
            self.fail('Creation should not raise exceptions')
        try:
            manager.remove('00000000')
        except ZCMError as e:
            self.fail('Creation should not raise exceptions')

        self.assertEqual(manager.path, Path(directory))
        self.assertEqual(manager.zfs, zfs)
        self.assertEqual(manager.active_clone, manager.clones[0])
        self.assertEqual(manager.next_id, '00000003')
        self.assertEqual(len(manager.older_clones), 0)
        self.assertEqual(len(manager.newer_clones), 1)

        filesystem = zfs
        path = Path(directory, '.clones')
        self.assertTrue(zfs_is_filesystem(filesystem))
        self.assertEqual(zfs_get(filesystem, 'mountpoint'), path)
        self.assertTrue(zfs_get(filesystem, 'mounted'))
        self.assertTrue(path.is_dir())

        id = '00000000'
        filesystem = '%s/%s' % (zfs, id)
        path = Path(directory, '.clones', id)
        self.assertFalse(zfs_exists(filesystem))
        self.assertFalse(path.exists())

        id = '00000001'
        filesystem = '%s/%s' % (zfs, id)
        path = Path(directory)
        self.assertEqual(manager.clones[0].id, id)
        self.assertEqual(manager.clones[0].zfs, filesystem)
        self.assertEqual(manager.clones[0].origin, '%s/%s@%s' % (zfs, '00000002', id))
        self.assertEqual(manager.clones[0].origin_id, '00000002')
        self.assertEqual(manager.clones[0].mountpoint, path)
        self.assertIsInstance(manager.clones[0].creation, datetime)
        self.assertTrue(zfs_is_filesystem(filesystem))
        self.assertEqual(zfs_get(filesystem, 'mountpoint'), path)
        self.assertTrue(zfs_get(filesystem, 'mounted'))
        self.assertTrue(path.is_dir())

        id = '00000002'
        filesystem = '%s/%s' % (zfs, id)
        path = Path(directory, '.clones', id)
        self.assertEqual(manager.clones[1].id, id)
        self.assertEqual(manager.clones[1].zfs, filesystem)
        self.assertIsNone(manager.clones[1].origin)
        self.assertIsNone(manager.clones[1].origin_id)
        self.assertEqual(manager.clones[1].mountpoint, path)
        self.assertIsInstance(manager.clones[1].creation, datetime)
        self.assertTrue(zfs_is_filesystem(filesystem))
        self.assertEqual(zfs_get(filesystem, 'mountpoint'), path)
        self.assertTrue(zfs_get(filesystem, 'mounted'))
        self.assertTrue(path.is_dir())

    def test_clone_options(self):
        clone = None
        try:
            manager = Manager(directory)
        except ZCMError as e:
            self.fail('Instantiation should not raise exceptions')
        try:
            manager.clone()
            manager.clone()
            manager.clone()
            manager.clone()
            manager.clone()
        except ZCMError as e:
            self.fail('Creation should not raise exceptions')
        with self.assertRaises(ZCMException):
            manager.clone(max_newer=5)
        with self.assertRaises(ZCMException):
            manager.clone(max_total=6)

        self.assertEqual(manager.path, Path(directory))
        self.assertEqual(manager.zfs, zfs)
        self.assertEqual(manager.active_clone, manager.clones[0])
        self.assertEqual(manager.next_id, '00000006')
        self.assertEqual(len(manager.older_clones), 0)
        self.assertEqual(len(manager.newer_clones), 5)

    def test_activate_options(self):
        clone = None
        try:
            manager = Manager(directory)
        except ZCMError as e:
            self.fail('Instantiation should not raise exceptions')
        try:
            manager.clone()
            manager.clone()
            manager.clone()
            manager.clone()
        except ZCMError as e:
            self.fail('Creation should not raise exceptions')
        with self.assertRaises(ZCMException):
            manager.activate('00000002', max_newer=1)
        with self.assertRaises(ZCMException):
            manager.activate('00000002', max_older=1)
        try:
            manager.activate('00000002', max_older=2, max_newer=2)
        except ZCMError as e:
            self.fail('Creation should not raise exceptions')

        self.assertEqual(manager.path, Path(directory))
        self.assertEqual(manager.zfs, zfs)
        self.assertEqual(manager.active_clone, manager.clones[2])
        self.assertEqual(manager.next_id, '00000005')
        self.assertEqual(len(manager.older_clones), 2)
        self.assertEqual(len(manager.newer_clones), 2)


if __name__ == '__main__':
    unittest.main()



# Copyright 2021, Guillermo Adrián Molina
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import logging

log = logging.getLogger(__name__)


class Clone:
    def __init__(self, id, zfs, origin, origin_id, mountpoint, creation, size):
        self.id = id
        self.zfs = zfs
        self.origin = origin
        self.origin_id = origin_id
        self.mountpoint = mountpoint
        self.creation = creation
        self.size = size

    def to_dictionary(self):
        return {
            'id': self.id,
            'zfs': self.zfs,
            'origin': self.origin,
            'origin_id': self.origin_id,
            'mountpoint': str(self.mountpoint),
            'creation': str(self.creation),
            'size':self.size
        }



# Copyright 2021, Guillermo Adrián Molina
#
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
from datetime import datetime
from pathlib import Path

from zcm.api.manager import Manager
from zcm.exceptions import ZCMError, ZCMException
from zcm.lib.zfs import ZFSError, zfs_create, zfs_destroy

zfs = 'rpool/my/cool/zfs/directory'
directory = '/my_cool_zfs_directory'


class TestAPI(unittest.TestCase):
    def test_initialize_existing_zfs(self):
        path = Path(directory)
        self.assertIsNotNone(zfs_create(zfs, mountpoint=directory, recursive=True))
        temp_dir = path.joinpath('my', 'cool', 'subdirectory')
        temp_dir.mkdir(parents=True)
        temp_file = temp_dir.joinpath('file.txt')
        with temp_file.open('w', encoding ='utf-8') as f:
            f.write('SOME TEXT')
        with self.assertRaises(ZCMError):
            Manager.initialize_manager(zfs, directory)
        try:
            zfs_destroy(zfs)
        except ZFSError:
            self.fail('zfs_destroy should not raise exceptions')
        self.assertFalse(temp_file.exists())
        path.rmdir()

    def test_migrate_existing_zfs(self):
        path = Path(directory)
        self.assertIsNotNone(zfs_create(zfs, mountpoint=directory, recursive=True))
        temp_dir = path.joinpath('my', 'cool', 'subdirectory')
        temp_dir.mkdir(parents=True)
        temp_file = temp_dir.joinpath('file.txt')
        with temp_file.open('w', encoding ='utf-8') as f:
            f.write('SOME TEXT')
        with self.assertRaises(ZCMError):
            Manager.initialize_manager(zfs, directory)
        with self.assertRaises(ZCMError):
            Manager.initialize_manager(zfs, directory, migrate='PATH')
        try:
            Manager.initialize_manager(zfs, directory, migrate='ZFS')
        except ZCMException as e:
            self.fail('Initialization should not raise exceptions')
        manager = None
        try:
            manager = Manager(zfs)
        except ZCMException as e:
            self.fail('Instantiation should not raise exceptions')
        self.assertTrue(temp_file.is_file())
        with temp_file.open('r', encoding ='utf-8') as f:
            self.assertEqual(f.read(), 'SOME TEXT')
        try:
            manager.destroy()
        except ZCMException as e:
            self.fail('Destroy should not raise exceptions')

    def test_migrate_existing_path(self):
        path = Path(directory)
        temp_dir = path.joinpath('my', 'cool', 'subdirectory')
        temp_dir.mkdir(parents=True)
        temp_file = temp_dir.joinpath('file.txt')
        with temp_file.open('w', encoding ='utf-8') as f:
            f.write('SOME TEXT')
        with self.assertRaises(ZCMError):
            Manager.initialize_manager(zfs, directory)
        with self.assertRaises(ZCMError):
            Manager.initialize_manager(zfs, directory)
        with self.assertRaises(ZCMError):
            Manager.initialize_manager(zfs, directory, migrate='ZFS')
        try:
            Manager.initialize_manager(zfs, directory, migrate='PATH')
        except ZCMException as e:
            self.fail('Initialization should not raise exceptions')
        manager = None
        try:
            manager = Manager(zfs)
        except ZCMException as e:
            self.fail('Instantiation should not raise exceptions')
        self.assertTrue(temp_file.is_file())
        with temp_file.open('r', encoding ='utf-8') as f:
            self.assertEqual(f.read(), 'SOME TEXT')
        try:
            manager.destroy()
        except ZCMException as e:
            self.fail('Destroy should not raise exceptions')

if __name__ == '__main__':
    unittest.main()



# Copyright 2021, Guillermo Adrián Molina
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import argparse
import json

from zcm.api.manager import Manager
from zcm.lib.helpers import check_one_or_more


class Clone:
    name = 'clone'
    aliases = []

    @staticmethod
    def init_parser(parent_subparsers):
        parent_parser = argparse.ArgumentParser(add_help=False)
        parser = parent_subparsers.add_parser(Clone.name,
                                              parents=[parent_parser],
                                              aliases=Clone.aliases,
                                              formatter_class=argparse.ArgumentDefaultsHelpFormatter,
                                              description='Create a new clone from the active clone',
                                              help='Create a new clone')
        parser.add_argument('-j', '--json',
                            action='store_true',
                            help='Output a JSON object')
        parser.add_argument('-m', '--max-newer',
                            type=check_one_or_more,
                            help='Do not create if there are <max-newer> newer clones')
        parser.add_argument('-t', '--max-total',
                            type=check_one_or_more,
                            help='Do not create if there are <max-total> clones')
        parser.add_argument('-a', '--auto-remove',
                            action='store_true',
                            help='Remove clones if maximum limits excedeed')
        parser.add_argument('path',
                            metavar='filesystem|path',
                            help='zfs filesystem or path of ZCM')

    def __init__(self, options):
        manager = Manager(options.path)
        clone = manager.clone(
            options.max_newer, options.max_total, options.auto_remove)
        if not options.quiet:
            if options.json:
                print(json.dumps(clone.to_dictionary(), indent=4))
            else:
                print('Created clone %s at path %s' %
                    (clone.id, clone.mountpoint))



# Copyright 2021, Guillermo Adrián Molina
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__version__ = "3.4.0"
AUTHOR = "Guillermo Adrián Molina"
AUTHOR_EMAIL = "guillermoadrianmolina@hotmail.com"
NAME = "zfs-clone-manager"
PACKAGE_URL = "http://www.github.com/guillermomolina/zfs-clone-manager"
KEYWORDS = "zfs"
DESCRIPTION = "Tool to manage ZFS clones with history metadata"
LICENSE = "LICENSE"



# Copyright 2021, Guillermo Adrián Molina
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse

from zcm.api import Manager
from zcm.lib.print import format_bytes, print_info, print_table


class Information:
    name = 'information'
    aliases = ['info']

    @staticmethod
    def init_parser(parent_subparsers):
        parent_parser = argparse.ArgumentParser(add_help=False)
        parser = parent_subparsers.add_parser(Information.name,
                                              parents=[parent_parser],
                                              aliases=Information.aliases,
                                              formatter_class=argparse.ArgumentDefaultsHelpFormatter,
                                              description='Show ZCM information',
                                              help='Show ZCM information')
        parser.add_argument('-t', '--table',
                            help='Show information as table',
                            action='store_true')
        parser.add_argument('path',
                            nargs='*',
                            metavar='filesystem|path',
                            help='zfs filesystem or path to show')

    def __init__(self, options):
        managers = []
        if options.path:
            managers = [ Manager(path) for path in options.path ]
        else:
            managers = Manager.get_managers()
        if options.table:
            table = []
            for manager in managers:
                table.append({
                    'path': manager.path,
                    'zfs': manager.zfs,
                    'size': format_bytes(manager.size),
                    'total': len(manager.clones),
                    'older': len(manager.older_clones),
                    'newer': len(manager.newer_clones),
                    'oldest_id': manager.clones[0].id,
                    'active_id': manager.active_clone.id,
                    'newest_id': manager.clones[-1].id,
                    'next_id': manager.next_id
                })
            print_table(table)
        else:
            for manager in managers:
                data = {
                    'Path': manager.path,
                    'Root ZFS': manager.zfs,
                    'Root ZFS size': format_bytes(manager.size),
                    'Total clone count': len(manager.clones),
                    'Older clone count': len(manager.older_clones),
                    'Newer clone count': len(manager.newer_clones),
                    'Oldest clone ID': manager.clones[0].id,
                    'Active clone ID': manager.active_clone.id,
                    'Newest clone ID': manager.clones[-1].id,
                    'Next clone ID': manager.next_id
                }
                print_info(data)
                print()



# Copyright 2021, Guillermo Adrián Molina
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import shutil
from pathlib import Path

from zcm.api.clone import Clone
from zcm.exceptions import ZCMError, ZCMException
from zcm.lib.helpers import copy_directory, id_generator
from zcm.lib.zfs import (ZFSError, zfs_clone, zfs_create, zfs_destroy,
                         zfs_inherit, zfs_list, zfs_mount, zfs_promote,
                         zfs_rename, zfs_set, zfs_snapshot, zfs_unmount)

log = logging.getLogger(__name__)


def get_zcm_for_path(path_str):
    path = Path(path_str).absolute()
    if not path.is_dir():
        return None
    absolute_path_str = str(path)
    zfs_list_output = zfs_list(absolute_path_str, zfs_type='filesystem', properties=[
                               'name', 'zfs_clone_manager:path', 'mountpoint'])
    if len(zfs_list_output) != 1:
        return None
    zfs = zfs_list_output[0]
    if str(zfs['zfs_clone_manager:path']) == absolute_path_str and str(zfs['mountpoint']) == absolute_path_str:
        splitted_name = zfs['name'].split('/')
        name = '/'.join(splitted_name[:-1])
        try:
            int(splitted_name[-1], base=16)
        except ValueError:
            return None
        return name


def snapshot_to_origin_id(snapshot):
    # snapshot -> rpool/zfsa/zfsb/00000004@00000005
    # snapshot.split('/') -> ['rpool','zfsa','zfsb','00000004@00000005']
    # snapshot.split('/')[-1] -> '00000004@00000005'
    # snapshot.split('/')[-1].split('@') -> ['00000004','00000005']
    # snapshot.split('/')[-1].split('@')[0] -> '00000004'
    if snapshot:
        return snapshot.split('/')[-1].split('@')[0]
    return None


class Manager:
    def __init__(self, zfs_or_path):
        self.zfs = None
        self.path = None
        self.clones = []
        self.older_clones = []
        self.newer_clones = []
        self.active_clone = None
        self.next_id = None
        self.size = None
        zfs = get_zcm_for_path(zfs_or_path)
        self.zfs = zfs_or_path if zfs is None else zfs
        self.load()

    @staticmethod
    def get_managers():
        zfs_list_output = zfs_list(
            properties=['name', 'zfs_clone_manager:path', 'mountpoint'])
        return [Manager(zfs['name'])
                for zfs in zfs_list_output
                if zfs['zfs_clone_manager:path'] is not None and
                zfs['mountpoint'] == Path(zfs['zfs_clone_manager:path'], '.clones')]

    @staticmethod
    def initialize_manager(zfs_str, path_str, migrate=None):
        path = Path(path_str)
        zfs_list_output = zfs_list(zfs_str, zfs_type='all', properties=[
            'name', 'type', 'zfs_clone_manager:path', 'origin', 'mountpoint'], recursive=True)
        if zfs_list_output:
            zfs = zfs_list_output[0]
            if zfs['zfs_clone_manager:path']:
                raise ZCMError(
                    'The ZFS %s is a ZCM manager, will not initialize' % zfs_str)
            if len(zfs_list_output) > 1:
                raise ZCMError(
                    'The ZFS %s has children, can not initialize ZCM with it' % zfs_str)
            if zfs['type'] != 'filesystem':
                raise ZCMError('The ZFS %s is of type %s, can not initialize ZCM with it' % (
                    zfs_str, zfs['type']))
            if migrate != 'ZFS':
                raise ZCMError(
                    'ZFS %s already exists, will not initialize a manager with it' % zfs_str)
            if zfs['mountpoint'] != path and path.exists():
                 raise ZCMError(
                    'Path %s already exists (and it is not the ZFS %s mountpoint, can not use it' % (path_str, zfs_str))
            random_id = id_generator()
            zfs_parts = zfs_str.split('/')[:-1]
            zfs_parts.append(random_id)
            random_zfs = '/'.join(zfs_parts)
            try:
                zfs_unmount(zfs_str)
                zfs_rename(zfs_str, random_zfs)
                zfs_create(zfs_str, zcm_path=path_str, recursive=True)
                zfs_unmount(zfs_str)
                zfs_rename(random_zfs, zfs_str + '/00000000')
                zfs_set(zfs_str + '/00000000', mountpoint=path)
                zfs_mount(zfs_str + '/00000000')
                zfs_set(zfs_str, mountpoint=Path(path, '.clones'))
                zfs_mount(zfs_str)
                log.info('Migrated ZFS %s at path %s to ZCM' % (zfs_str, path_str))
            except ZFSError as e:
                raise ZCMError(e.message)            
            return
        original_path = None 
        if path.exists():
            if migrate != 'PATH':
                raise ZCMError(
                    'Path %s already exists, will not initialize a manager' % path_str)
            random_id = id_generator()
            original_path = Path(path.parents[0], random_id)
            path.rename(original_path) 
        try:
            zfs_create(zfs_str, zcm_path=path_str, recursive=True)
            zfs_unmount(zfs_str)
            zfs_create('00000000', zfs_str, mountpoint=path)
            zfs_set(zfs_str, mountpoint=Path(path, '.clones'))
            zfs_mount(zfs_str)
            log.info('Created ZCM %s at path %s' % (zfs_str, path_str))
        except ZFSError as e:
            raise ZCMError(e.message)
        if original_path:
            if copy_directory(original_path, path) != 0:
                raise ZCMError('Could not move content of original directory, kept at %s' % original_path)
            shutil.rmtree(original_path)
            log.info('Moved content of path %s to clone' % path_str)
           

    def load(self):
        if not isinstance(self.zfs, str):
            raise ZCMError(
                'The name property is invalid: ' + str(self.zfs))
        self.path = None
        self.clones = []
        self.older_clones = []
        self.newer_clones = []
        self.active_clone = None
        self.next_id = None
        self.size = None
        last_id = 0
        zfs_list_output = zfs_list(self.zfs, zfs_type='filesystem', properties=[
            'name', 'zfs_clone_manager:path', 'origin', 'mountpoint',
            'creation', 'used'], recursive=True)
        if not zfs_list_output:
            raise ZCMError(
                'There is no ZCM manager at %s' % self.zfs)
        for zfs in zfs_list_output:
            if self.path is None:
                self.zfs = zfs['name']
                if zfs['zfs_clone_manager:path'] is None:
                    raise ZCMError(
                        'The ZFS %s is not a valid ZCM manager' % zfs['name'])
                self.path = zfs['zfs_clone_manager:path']
                self.size = zfs['used']
            else:
                splitted_name = zfs['name'].split('/')
                name = '/'.join(splitted_name[:-1])
                id = splitted_name[-1]
                if name != self.zfs:
                    raise ZCMError(
                        'The ZFS %s is not a valid ZCM clone' % zfs['name'])
                try:
                    last_id = max(last_id, int(id, base=16))
                except ValueError:
                    raise ZCMError(
                        'The ZFS %s is not a valid ZCM clone' % zfs['name'])
                origin_id = snapshot_to_origin_id(zfs['origin'])
                clone = Clone(id, zfs['name'], zfs['origin'], origin_id,
                              zfs['mountpoint'], zfs['creation'], zfs['used'])
                if not isinstance(self.path, Path) or not self.path.is_dir():
                    raise ZCMError(
                        'The path property is invalid: ' + self.path)
                if zfs['mountpoint'] == self.path:
                    self.active_clone = clone
                else:
                    if self.active_clone:
                        self.newer_clones.append(clone)
                    else:
                        self.older_clones.append(clone)
                self.clones.append(clone)
        self.next_id = format(last_id + 1, '08x')

    def clone(self, max_newer=None, max_total=None, auto_remove=False):
        if not self.active_clone:
            raise ZCMError('There is no active clone, activate one first')
        if not auto_remove and max_newer is not None and len(self.newer_clones) >= max_newer:
            raise ZCMException(
                'There are already %d newer clones, can not create another' % len(self.newer_clones))
        if not auto_remove and max_total is not None and len(self.clones) >= max_total:
            raise ZCMException(
                'There are already %d clones, can not create another' % len(self.clones))
        id = self.next_id
        try:         
            snapshot = zfs_snapshot(id, self.active_clone.zfs)
            zfs = zfs_clone(self.zfs + '/' + id, snapshot)
        except ZFSError as e:
            raise ZCMError(e.message)            
        self.load()
        clone = self.get_clone(id)
        log.info('Created clone ' + clone.id)
        self.auto_remove(max_newer=max_newer, max_total=max_total)
        return clone

    def auto_remove(self, max_newer=None, max_older=None, max_total=None):
        while max_older is not None and len(self.older_clones) > max_older:
            self.remove(self.older_clones[0].id)
        while max_newer is not None and len(self.newer_clones) > max_newer:
            self.remove(self.newer_clones[0].id)
        while max_total is not None and len(self.clones) > max_total:
            if self.older_clones:
                self.remove(self.older_clones[0].id)
            elif self.newer_clones:
                self.remove(self.newer_clones[0].id)
            else:
                raise ZCMError(
                    'There are no more clones to remove in order to satisfy max limit of ' + max_total)

    def get_clone(self, id):
        for clone in self.clones:
            if clone.id == id:
                return clone
        raise ZCMError('There is no clone with id ' + id)

    def unmount(self):
        try:
            for clone in self.clones:
                if clone != self.active_clone:
                    zfs_unmount(clone.zfs)
            zfs_unmount(self.zfs)
            if self.active_clone is not None:
                zfs_unmount(self.active_clone.zfs)
        except ZFSError as e:
            # at lest one unmount failed, remount all and fail
            self.mount()
            raise ZCMError(e.message)

    def mount(self):
        if not self.active_clone:
            raise ZCMError('There is no active clone, activate one first')
        try:
            zfs_mount(self.active_clone.zfs)
            zfs_mount(self.zfs)
            for clone in self.clones:
                if clone != self.active_clone:
                    zfs_mount(clone.zfs)
        except ZFSError as e:
            raise ZCMError(e.message)

    def activate(self, id, max_newer=None, max_older=None, max_total=None, auto_remove=False):
        next_active = self.get_clone(id)
        if next_active == self.active_clone:
            raise ZCMException('Manager %s already active' % id)
        if not auto_remove and (max_newer is not None or max_older is not None):
            newer_count = 0
            older_count = 0
            has_reach_active = False
            for clone in self.clones:
                if clone == next_active:
                    has_reach_active = True
                else:
                    if has_reach_active:
                        newer_count += 1
                    else:
                        older_count += 1
            if not auto_remove and max_newer is not None and newer_count > max_newer:
                raise ZCMException(
                    'Command denied, Activating %s violates the maximum number of newer clones (%d/%d)'
                    % (id, newer_count, max_newer))
            if not auto_remove and max_older is not None and older_count > max_older:
                raise ZCMException(
                    'Command denied, Activating %s violates the maximum number of older clones (%d/%d)'
                    % (id, older_count, max_older))

        self.unmount()
        try:
            if self.active_clone is not None:
                zfs_inherit(self.active_clone.zfs, 'mountpoint')
            zfs_set(next_active.zfs, mountpoint=self.path)
        except ZFSError as e:
            raise ZCMError(e.message)
        self.active_clone = next_active
        self.mount()

        log.info('Activated clone ' + id)
        self.load()
        self.auto_remove(max_newer=max_newer,
                         max_older=max_older, max_total=max_total)
        return next_active

    def find_clones_with_origin(self, id):
        clones = []
        for clone in self.clones:
            if clone.origin_id == id:
                clones.append(clone)
        return clones

    def remove(self, id):
        clone = self.get_clone(id)
        if clone == self.active_clone:
            raise ZCMError(
                'Manager with id %s is active, can not remove' % id)
        clones = self.find_clones_with_origin(id)
        promoted = None
        if clones:
            promoted = clones[-1]
            zfs_promote(promoted.zfs)
        try:
            zfs_destroy(clone.zfs)
            if clone.origin:
                zfs_destroy(clone.origin)
            if promoted:
                zfs_destroy('%s@%s' % (promoted.zfs, promoted.id))
            log.info('Removed clone ' + clone.id)
            self.load()
        except ZFSError as e:
            raise ZCMError(e.message)            
        

    def destroy(self):
        try:
            self.unmount()
            zfs_destroy(self.zfs, recursive=True)
            self.path.rmdir()
        except ZFSError as e:
            raise ZCMError(e.message)            
        except OSError as e:
            raise ZCMError('Could not destroy path ' + self.path)

    def to_dictionary(self):
        return {
            'zfs': self.zfs,
            'path': str(self.path),
            'size': self.size,
            'clones': [ clone.id for clone in self.clones ],
            'older_clones': [ clone.id for clone in self.older_clones ],
            'newer_clones': [ clone.id for clone in self.newer_clones ],
            'active_clone': self.active_clone.id,
            'next_id': self.next_id
        }



# Copyright 2021, Guillermo Adrián Molina
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from setuptools import setup, find_packages
import os

################################################################################
# HELPER FUNCTIONS #############################################################
################################################################################

def get_lookup():
    '''get version by way of the version file
    '''
    lookup = dict()
    version_file = os.path.join('zcm', 'version.py')
    with open(version_file) as filey:
        exec(filey.read(), lookup)
    return lookup

def get_install_requirements():
    with open('requirements.txt') as f:
        requirements = f.read().splitlines()
        return requirements

# Make sure everything is relative to setup.py
install_path = os.path.dirname(os.path.abspath(__file__)) 
os.chdir(install_path)

# Get version information from the lookup
lookup = get_lookup()
VERSION = lookup['__version__']
NAME = lookup['NAME']
AUTHOR = lookup['AUTHOR']
AUTHOR_EMAIL = lookup['AUTHOR_EMAIL']
PACKAGE_URL = lookup['PACKAGE_URL']
KEYWORDS = lookup['KEYWORDS']
DESCRIPTION = lookup['DESCRIPTION']
LICENSE = lookup['LICENSE']
with open('README.md') as readme:
    LONG_DESCRIPTION = readme.read()

################################################################################
# MAIN #########################################################################
################################################################################

if __name__ == "__main__":

    INSTALL_REQUIRES = get_install_requirements()
    #DEV_REQUIRES = get_dev_requirements()
    #TESTS_REQUIRES = get_test_requirements()

    setup(
        name=NAME,
        version=VERSION,
        author=AUTHOR,
        author_email=AUTHOR_EMAIL,
        maintainer=AUTHOR,
        maintainer_email=AUTHOR_EMAIL,
        packages=find_packages(), 
        include_package_data=True,
        zip_safe=False,
        url=PACKAGE_URL,
        license=LICENSE,
        description=DESCRIPTION,
        long_description=LONG_DESCRIPTION,
        long_description_content_type="text/markdown",
        keywords=KEYWORDS,
        test_suite="tests",
        #tests_require=TESTS_REQUIRES,
        install_requires=INSTALL_REQUIRES,
        entry_points={
            'console_scripts': [
                'zcm = zcm.cli.main:main'
            ]
        },
        classifiers=[
            'Development Status :: 5 - Production/Stable',
            'Intended Audience :: Developers',
            'Intended Audience :: System Administrators',
            'License :: OSI Approved :: Apache Software License',
            'Operating System :: Unix',
            'Programming Language :: Python :: 3.7',
            'Topic :: Software Development :: Libraries :: Python Modules',
            'Topic :: System :: Operating System',
        ]
    )



# Copyright 2021, Guillermo Adrián Molina
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import argparse

from zcm.api.manager import Manager
from zcm.lib.print import print_table
from zcm.lib.zfs import zfs_diff


class Difference:
    name = 'difference'
    aliases = ['diff']

    @staticmethod
    def init_parser(parent_subparsers):
        parent_parser = argparse.ArgumentParser(add_help=False)
        parser = parent_subparsers.add_parser(Difference.name,
                                              parents=[parent_parser],
                                              aliases=Difference.aliases,
                                              formatter_class=argparse.ArgumentDefaultsHelpFormatter,
                                              description='Gives a high-level description of the differences between\
                                                  a clone and its origin',
                                              help='Differencies between a clone and its origin')
        parser.add_argument('-T', '--no-trunc',
                            help='Don\'t truncate output',
                            action='store_true')
        parser.add_argument('-H', '--no-header',
                            help='Don\'t show header line(s)',
                            action='store_true')
        parser.add_argument('-P', '--page-size',
                            help='If list is longer than <page-size>, ask for more to continue in <page-size> intervals.\
                                Enter 0 to avoid pagination',
                            type=int,
                            default=25)
        parser.add_argument('path',
                            metavar='filesystem|path',
                            help='zfs filesystem or path of ZCM')
        parser.add_argument('id',
                            nargs='?',
                            help='clone id',
                            default='active')

    def __init__(self, options):
        manager = Manager(options.path)
        id = manager.active_clone.id if options.id == 'active' else options.id
        clone = manager.get_clone(id)
        table = zfs_diff(clone.zfs, clone.origin, include_file_types=True)
        print_table(table, header=(not options.no_header), truncate=(
            not options.no_trunc), page_size=options.page_size)



# Copyright 2021, Guillermo Adrián Molina
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import argparse

from zcm.api.manager import Manager


class Initialize:
    name = 'initialize'
    aliases = ['init']

    @staticmethod
    def init_parser(parent_subparsers):
        parent_parser = argparse.ArgumentParser(add_help=False)
        parser = parent_subparsers.add_parser(Initialize.name,
                                              parents=[parent_parser],
                                              aliases=Initialize.aliases,
                                              formatter_class=argparse.ArgumentDefaultsHelpFormatter,
                                              description='Init ZCM on specified ZFS',
                                              help='Init ZCM on specified ZFS')
        migrate_parser_group = parser.add_mutually_exclusive_group()
        migrate_parser_group.add_argument('-m', '--migrate-zfs',
                            help='Migrate existing ZFS',
                            action='store_true')
        migrate_parser_group.add_argument('-M', '--migrate-path',
                            help='Migrate existing path',
                            action='store_true')
        parser.add_argument('zfs',
                            metavar='filesystem',
                            help='root ZFS filesystem for manager')
        parser.add_argument('path',
                            help='path to use for active clone')

    def __init__(self, options):
        migrate = None
        if options.migrate_zfs:
            migrate = 'ZFS'
        if options.migrate_path:
            migrate = 'PATH'
        Manager.initialize_manager(options.zfs, options.path, migrate)
        if not options.quiet:
            print('ZCM initialized ZFS %s at path %s' %
                  (options.zfs, options.path))
